#ifndef MULTI_INDEX_ALGORITHMS_HPP
#define MULTI_INDEX_ALGORITHMS_HPP

#if defined(_MSC_VER)&&(_MSC_VER>=1200)
#pragma once
#endif

#include "custom_partition.hpp"

#include <boost/multi_index/ordered_index.hpp>

namespace boost {

/* Template functions lower_bound, upper_bound, equal_range and binary_search
expressed in terms of partition_point, special version for iterators ranging in
multi_index_container's index

NodeBase is a base type of the multi_index_container's node (internal data
structure used to store elements). */

#if defined(BOOST_MULTI_INDEX_ENABLE_SAFE_MODE)
BOOST_CUSTOM_PARTITION(
  ((typename)(NodeBase))((typename)(OrderedIndex)),
  boost::multi_index::safe_mode::safe_iterator<
    boost::multi_index::detail::bidir_node_iterator<boost::multi_index::detail::ordered_index_node<NodeBase> >
    BOOST_PP_COMMA()
    OrderedIndex
  >,
  itBegin, itEnd, UnaryPred, pred
)
#else
BOOST_CUSTOM_PARTITION(
  ((typename)(NodeBase)),
  boost::multi_index::detail::bidir_node_iterator<boost::multi_index::detail::ordered_index_node<NodeBase> >,
  itBegin, itEnd, UnaryPred, pred
)
#endif
{
  typedef boost::multi_index::detail::ordered_index_node<NodeBase> node_type;

  if(itBegin==itEnd) return itEnd;

  /* find special nodes of the underlying internal tree: root and header;
  header points to the end of the index and is used here instead of end() */
  node_type* root=itBegin.get_node();
  node_type* header=
    node_type::from_impl(root->parent());
  node_type* next=
    node_type::from_impl(header->parent());
  while(root!=next) {
    root=header;
    header=next;
    next=node_type::from_impl(next->parent());
  }

  /* find the node of the partition point, using binary search */
  node_type *top = root;
  node_type *i = header;
  while(top) {
    if(pred(top->value())){
      i=top;
      top=node_type::from_impl(top->left());
    }
    else top=node_type::from_impl(top->right());
  }

  /* find its projection on the interval */
  if(i==header)
    return itEnd;
  if((itEnd.get_node()!=header) && (!pred(*itEnd) && pred(i->value())))
      return itEnd;
  if(pred(*itBegin) || !pred(i->value()))
    return itBegin;
#if defined(BOOST_MULTI_INDEX_ENABLE_SAFE_MODE)
  return boost::multi_index::safe_mode::safe_iterator<boost::multi_index::detail::bidir_node_iterator<boost::multi_index::detail::ordered_index_node<NodeBase> >, OrderedIndex>(i, const_cast<OrderedIndex*>(itBegin.owner()));
#else
  return boost::multi_index::detail::bidir_node_iterator<boost::multi_index::detail::ordered_index_node<NodeBase> >(i);
#endif
}

}; /* namespace */


#endif /* MULTI_INDEX_ALGORITHMS_HPP */
